#include "OgrePlatformQueryManager.h"
#include "OcclusionQuery.h"
#include <OgreSceneManager.h>
#include <OgreLogManager.h>

#include <vector>


namespace Ogre {
//----------------------------------------------------------------------- 
PlatformQueryManager::PlatformQueryManager(PlatformHierarchyInterface *hierarchyInterface, Viewport *vp):
QueryManager(hierarchyInterface), mViewport(vp)
{
}
//-----------------------------------------------------------------------
bool PlatformQueryManager::ShootRay(const Ray &ray, std::vector<Mesh *> *visibleMeshes, bool isGlobalLine)
{
    // run OGRE ray shooting query
    return false;
}
//-----------------------------------------------------------------------
void PlatformQueryManager::ComputeCameraVisibility(const Camera &camera,
			    InfoContainer<GtpVisibility::NodeInfo> *visibleNodes,
			    InfoContainer<GtpVisibility::MeshInfo> *visibleGeometry,
			    bool relativeVisibility)
{
	if (relativeVisibility)
		LogManager::getSingleton().logMessage("relative");
	else
		LogManager::getSingleton().logMessage("absolute");

	// we need access to the scene manager and the rendersystem
	PlatformHierarchyInterface *pfHierarchyInterface = 
		dynamic_cast<PlatformHierarchyInterface *>(mHierarchyInterface);

	//-- Render scene to get conservative visibility and fill depth buffer 

	// can do const_cast because Camera no changed in renderScene
	Camera *pCam = const_cast<Camera *>(&camera); 

	pfHierarchyInterface->GetSceneManager()->_renderScene(pCam, mViewport, false);

	/* 
		Test again to get exact visibility with regard to the current camera.
		Issue all queries first to avoid starvation & stalls.
	*/
	GtpVisibility::QueryList queryList[2];
	
	// get rendered hierarchy nodes
	GtpVisibility::HierarchyNodeList *nodeList = mHierarchyInterface->GetRenderedNodes();
	// vector for storing entities of meshes
	GtpVisibility::GeometryList geometryList;
	
	GtpVisibility::HierarchyNodeList::const_iterator nodeIt, nodeIt_end = nodeList->end();
	// geometry list has still do be built
	GtpVisibility::GeometryList::iterator geometryIt, geometryIt_end;

	// for relative visibility we need 2 rendering passes
	int n = relativeVisibility ? 2 : 1;
	
	for (int i=0; i<n; ++i)
	{
		//-- queries for hierarchy nodes
		for (nodeIt = nodeList->begin(); nodeIt != nodeIt_end; ++nodeIt)
		{
			// TODO: DELETE QUERIES FROM PREVIOUS RENDER
			queryList[i].push_back(mHierarchyInterface->IssueOcclusionQuery(*nodeIt, false));
			
			// store geometry of nodes in geometry list
			mHierarchyInterface->GetGeometry(*nodeIt, &geometryList, false);
		}

		geometryIt_end = geometryList.end();

		//-- add queries for geometry
		for (geometryIt = geometryList.begin(); geometryIt != geometryIt_end; ++geometryIt)
		{
			queryList[i].push_back(mHierarchyInterface->IssueOcclusionQuery(*geometryIt));
		}

		// --- relative visiblity
		/* 
			1) get visible pixels count of objects
			2) clear frame buffer
			3) get projected visible pixels count:
			   test all objects again without depth write (set as option in scene manager)
			4) calculate ratio between visible vs. projected pixels
		*/
		if (i == n - 2) // if relativeVisibility, between first and second loop
		{
			pfHierarchyInterface->GetRenderSystem()->clearFrameBuffer(FBT_COLOUR | FBT_DEPTH);
		}

		// to obtain the correct number of projected pixels, depth write must be disabled
		bool enableDepthCheck = (i != n - 2);

		// this option must be provided by the scene manager
		pfHierarchyInterface->GetSceneManager()->setOption("DepthCheck", &enableDepthCheck);
	}


	//---- collect results
	unsigned int visiblePixels;

	GtpVisibility::QueryList::iterator absQueryIt, relQueryIt;

	absQueryIt = queryList[0].begin();
	relQueryIt = queryList[1].begin();
	
	for (nodeIt = nodeList->begin(); nodeIt != nodeIt_end; ++nodeIt)
	{
		(*absQueryIt)->GetQueryResult(visiblePixels, true);

		float vis = (float)visiblePixels;

		if (relativeVisibility)
		{
			(*relQueryIt)->GetQueryResult(visiblePixels, true);
			std::stringstream d; d << "relativ " << visiblePixels;
			LogManager::getSingleton().logMessage(d.str());

			if (visiblePixels > 0)
			{				
				vis /= (float) visiblePixels;
			}
			++relQueryIt;
		}

		++absQueryIt;

		// leave nodes with visibilty 0 in queue: 
		// happens if node is intersected by near plane
		visibleNodes->push_back(GtpVisibility::NodeInfo(*nodeIt, vis));
	}

	//---- queries for geometry
	geometryIt_end = geometryList.end();
	
	/*for (geometryIt = geometryList.begin(); geometryIt != geometryIt_end; ++geometryIt)
	{
		(*absQueryIt)->GetQueryResult(visiblePixels, true);

		float vis = (float)visiblePixels;
		bool isVisible = visiblePixels > 0;
		
		if (relativeVisibility)
		{
			(*relQueryIt)->GetQueryResult(visiblePixels, true);

			if (visiblePixels)
			{
				vis /= (float) visiblePixels;
			}
			++relQueryIt;
		}

		++absQueryIt;

		// approximate depth ordering during rendering => 
		// geometry maybe occluded
		if (isVisible)
		{
			visibleGeometry->push_back(GtpVisibility::MeshInfo(*geometryIt, vis));
		}
	}*/

}
//-----------------------------------------------------------------------
void PlatformQueryManager::ComputeFromPointVisibility(const Vector3 &point,
			       InfoContainer<GtpVisibility::NodeInfo> *visibleNodes,
			       InfoContainer<GtpVisibility::MeshInfo> *visibleGeometry,
			       bool relativeVisibility)
{
	 visibleNodes->push_back(GtpVisibility::NodeInfo(mHierarchyInterface->GetSceneRoot(), 1.0f));
}
//-----------------------------------------------------------------------
void PlatformQueryManager::SetViewport(Viewport *vp)
{
	mViewport = vp;
}
} // namespace Ogre
